package kraft.monitor.adapter;

import kraft.monitor.display.Message;

import kraft.TermInString;



/** Represents a kraft_msg/2 term as an object that is 
  * recognised as a Message by the Monitor.
  *
  * Parses a (string) term structure into context and core as shown
  * below:
  * <pre>
  * kraft_msg( context , Core), Core= ccql(Perf, [sender:agent1,...])
  * </pre>
  *
  * The core of the message is parsed as a CCQL object.
  *
  *@author Ted Francis
  *@see CCQL
  *@see kraft.monitor.display.Message
  *
  */

//=========================================
public class Kraft_msg implements Message {
//=========================================

  /** Field extracted from the context of the message
    */
  public String sender = "", receiver = "", time_stamp="";

  
  private CCQL core;
  private String context;



  /** Construct a new kraft_msg object from the given
    * string. This must be a prolog term structure represented
    * as string.
    *
    *@exception Exception If the string is not valid kraft_msg term.
    */
  public Kraft_msg(String str) throws Exception {
    this(new TermInString(str));
  }


  /** Construct a new kraft_msg object from the given
    * Term 
    *
    *@exception Exception If the term is not a valid kraft_msg term
    */
  public Kraft_msg(TermInString term) throws Exception {
    String err_msg = null;

    try{ 
      //make sure it is a valid term structure
      
      if(!term.functor.equals("kraft_msg") || term.arity != 2) 
	err_msg = ("invalid kraft_msg term: " +term.term);
	
      TermInString cont = new TermInString(term.argument(1));
      if(!cont.functor.equals("context") ||
	 cont.arity != 5) 
	err_msg = ("invalid context part of kraft_msg: "+ term.argument(1));

      //maybe have err-msg for time_stamp also..?? risky as Linja client
      // might not have one.

      core = new CCQL(term.argument(2));
      
      sender     = cont.argument(3);
      receiver   = cont.argument(4);
      time_stamp = cont.argument(5);
      context    = cont.term; //store it as a string

    }catch(Exception e) {
      throw new Exception("Error parsing kraft_msg term:\n"+
			  "--> term = " + term.term + "\n" +
			  "--> error= " + e.getClass().getName() );
      
    }

    //System.out.println("Parsed Kraft_msg term OK --> ");

  }//end constructor

  
  /** Return the Core of the message as CCQL object
    */
  public CCQL getCore() {
    return core;
  }


  /** Return the context of the message as a string
    */
  public String getContext(){
    return context;
  }



  //methods to implement the Message interface


  /** Returns the performative of the CCQL part of the message
    */
  public String getCommandString() {
    return core.performative;
  }


  /** Returns the whole kraft_msg as a simple string*/
  public String toString() {
    return ("kraft_msg(\n   "+context+",\n   "+core+")");
  }



  /** Format the message into a string for displaying
    *
    *@param detail if 1, displays CCQL.getDisplayString().
    *              if 2, displays context + CCQL.getDisplayString().
    *
    *@see CCQL#getDisplayString()
    */
  public String getDisplayString(int detail) {

    if(detail <= 0) detail = 1;
    if(detail >  2) detail = 2;

    switch(detail) {
    case 1: return core.getDisplayString(1);

    case 2: default:     
    return ("kraft_msg( "+context+ ","+
	    "\n\n"+                //split in windows happens here
	    core.getDisplayString(1)
	    +"\n)");

    }//end switch
  }//end method



  /** Whether obj represents the same message as this one
    */
  public boolean equals(Object obj) {
    if (obj instanceof Kraft_msg) {
      Kraft_msg other = (Kraft_msg) obj;
      
      if(core.equals(other.core) &&
	 context.equals(other.context))
	return true;
    }
    return false;
  }//end equals      
  
}//end class Kraft_msg
