package kraft.monitor.adapter;


import Linja.*;
import kraft.monitor.display.*;
import kraft.monitor.tools.*;

import java.util.Vector;
import java.net.*;
import java.io.*;
import java.util.*;

import kraft.TermInString;


/** Deals with receiving messages from a Linda server on behalf
  * of the Monitor. 
  *
  * On startup, connects to a Linda server, and "out"s predicates
  * to inform the Linda server(s) being monitored to make duplicates.
  *
  * for local site:    monitor(<mon_name>)
  * for remote site:   remote_monitor(on, <r_site>, <mon_name>)
  *
  * Converts the raw strings received from messages (representing Prolog
  * terms) into Kraft_msg objects, and registers KraftAgents when
  * needed to represent the sender/receiver of messages.
  *
  *It passes these to any MonitorServer, using the standard interface
  *methods. Currently, the MonitorGUI is the one receiving these calls.
  *
  *@author Ted Francis
  *@see Kraft_msg
  *@see KraftAgent
  *@see MonitorGUI
  */

//=========================
public class LinjaAdapter {
//=========================

  private MonitorServer monitor;             //the monitor I will connect to

  private String my_name                 = "";          // my kraft name
  private String local_monitor_predicate = "";

  private String           linda_host = "";		//linda-server host
  private int              linda_port = 0;		//linda-server port
  private LindaClient      linda      = null;
  private LindaMsgReceiver receiver   = null;


  // constructors


  /** Start up and prepare to send info to 'mon'
    * Should be followed by a call to connectToLinda()
    */
  public LinjaAdapter(MonitorServer mon){
    this.monitor = mon;
  }


  /** Start up and prepare to send info to 'mon'.
    * Connects to the linda server running on 'lhost:lport' using
    * name "name" for the monitor
    */
  public LinjaAdapter(MonitorServer mon, String lhost, int lport, String name){	
    this(mon);
    connectToLinda(lhost,lport,name);
  }

  /** Start up and prepare to send info to 'mon'.
    * Connects to the linda server running on 'lhost:lport' using
    * name "krl( site, name )" for the monitor 
    */
  public LinjaAdapter(MonitorServer mon, String lhost, 
		      int lport, String name, String site)
    {	
    this(mon);
    connectToLinda(lhost,lport,name,site);
    }


  /** Connect to the linda server at "lhost:lport", using the 
    * name "krl( site, name)" for the monitor
    */
  public void connectToLinda(String lhost, int lport, 
			     String name, String site) {
    if(site==null || site.equals(""))
      connectToLinda(lhost,lport,name);
    else
      connectToLinda(lhost,lport,"krl("+site+","+name+")");
  }


  /** Connect to the linda server at "lhost:lport" using the 
    * name 'name' for the monitor
    */
  public void connectToLinda(String lhost, int lport, String name){

    if(name!=null &&
       my_name.equals(no_spaces(name)) &&
       linda_host.equals(no_spaces(lhost)) &&
       linda_port == lport)
      return;

      my_name     = no_spaces(name);
      linda_host  = lhost;
      linda_port  = lport;

      local_monitor_predicate = ("monitor(" + my_name + ")");

      //disconnect from old linda_server, if we were connected
      if(linda !=null) {
	stop();
      }
	
      try{
	
	linda = new LindaClient(linda_host,linda_port);
	linda.connect();
	System.out.println("Connect to linda: "+linda_host+":"+linda_port);
	clean_mess(linda);
	linda.out(local_monitor_predicate);
	linda.disconnect();
	System.out.println("Out to linda: "+local_monitor_predicate);
      }
      catch(Exception e) {
	System.err.println("Error connecting to Linda "+linda_host+
			   ":"+linda_port+"\n--->"+e);
	return;
      }
      
      //Set up a listening thread to pick up messages from Linda
      // give it a handle to the MonitorServer so it can register
      // information with it.

      receiver = new LindaMsgReceiver(this.monitor, linda_host, 
				      linda_port, my_name);
      receiver.start();

  }//end connectToLinda


  /**Take out all spaces from the string*/
  //-----------------------------------
  private String no_spaces(String str){
  //-----------------------------------

    StringTokenizer tok = new StringTokenizer(str," ",false);
    String rtn = "";
    while(tok.hasMoreTokens())
      rtn += tok.nextToken();
    return rtn;
  }



  private Vector remote_sites = null;


  /** Try to monitor the given remote sites.
    * Sends remote_monitor(on,_,_) to the given sites.
    * Also sends remote_monitor(off,_,_) to the previously
    * monitored sites.
    */
  //----------------------------------------------
  public void monitorRemoteSites(String[] new_sites) {
  //----------------------------------------------

    if(new_sites==null || new_sites.length<=0) {
      return;
    }
    
    if (remote_sites==null) 
      remote_sites = new Vector(new_sites.length);

    try {
      linda.connect();

      //Send remote_monitor(off,) predicates to previously monitored sites

      for(int i=0; i<remote_sites.size(); i++) {
	String s = (String)remote_sites.elementAt(i);
	linda.out( remote_monitor_predicate(false, s) );
      }
      remote_sites.removeAllElements();

      
      // Now send remote_monitor(on,) predicates to all the new_sites

      for(int i=0; i < new_sites.length; i++) {
	if(new_sites[i]!=null) {
	  String curr_site = no_spaces(new_sites[i]);

	  linda.out( remote_monitor_predicate(true, curr_site) );
	  System.out.println("Out to linda: "+
			     remote_monitor_predicate(true, curr_site));
	  remote_sites.addElement(curr_site);
	}
      }
      linda.disconnect();
    }
    catch(IOException e) {
      System.err.println("Error: problems sending remote_monitor pred's");
      e.printStackTrace();
    }

  }
  
  //---------------------------------------------------------------------
  private String remote_monitor_predicate(boolean on, String site_name) {
  //---------------------------------------------------------------------
    return ("remote_monitor("+
	    ((on)?"on":"off")+
	    ","+site_name+","+my_name+")");
  }


  /** Stop the linda adapter from receiving messages, and instruct
    * the monitored site(s) to stop sending duplicates
    */
  //------------------
  public void stop() {
  //------------------

    Enumeration enum;
    String msg;

    //Make local Linda stop sending duplicates (and ASK remote Lindas to)
    //-------------------------------------------------------------------
    try{
      linda.connect();

      clean_mess(linda); 

      if(remote_sites!=null) {
	enum = remote_sites.elements();
	while(enum.hasMoreElements()){
	  String pred = remote_monitor_predicate(false, 
						 (String)enum.nextElement());
	  linda.out(pred);
	  System.out.println("Out to linda: "+pred);
	}
      }
      linda.disconnect();

      //stop listening for messages -- also cleans up a bit
      receiver.stop_listening();
    }
    catch(IOException e) {}

    receiver = null;
    linda    = null;
  }

  /** Clean up the monitor/1 predicate left earlier.*/
  //-------------------------------------------------------------
  private void clean_mess(LindaClient linda) throws IOException {
  //-------------------------------------------------------------
    String msg;

    if(linda!=null) {
      linda.set_timeout(500);
      msg = linda.in_noblock(local_monitor_predicate); 
      if(msg!=null)
	System.out.println("In from linda: "+msg);
    }
  }


}//end class LinjaAdapter



/** Used by the LinjaAdapter to receive messages on a 
  * separate thread.
  */
//=====================================
class LindaMsgReceiver extends Thread {
//=====================================

  Vector names;                 //names of known 'agents'
  MessageHashtable messages;    // messages stored to check for duplicates

  MonitorServer monitor;
  LindaClient linda=null;		//linda client for connection
  String[] mask;			//the 'in' mask
  String   my_name;
  String   monitor_pred;
  

  //----------------------------------------
  public LindaMsgReceiver(MonitorServer mon,
			  String host, 
			  int port,
			  String name) {
  //----------------------------------------
 
     super();
      linda   =new LindaClient(host,port);
      monitor = mon;
      my_name = name;
      names    = new Vector(5,5);
      messages = new MessageHashtable();


      mask = new String[3];
      mask[0]="monitor("+my_name+",_)"; //being used at the moment
                                        //while Linda adjusted to use Messaging
                                        // API to make 'forward's
      mask[1]="kraft_msg(context(1,_,_,"+name+",_),_)";
      mask[2]="kraft_msg(context(1,_,_,krl(_,"+name+"),_),_)";

  }//end constructor



  /** Start listening for messages*/
  //-----------------
  public void run() {
  //-----------------
    
    String message;
    
    try {
      linda.connect();
      cleanLinda();  //in case of old messages left behind from before
    }
    catch (SocketException e)
      {
	System.err.println("ERROR: cannot connect to Linda");
	System.err.println("Shutting down message receiver...");
	stop_listening();				//stop thread
      }
    catch (UnknownHostException e)
      {
	System.err.println("ERROR: unknown Linda host");
	System.err.println("Shutting down message receiver...");
	stop_listening();
      }
    catch (IOException e)
      {
	System.err.println("ERROR: I/O error to Linda");
	System.err.println("Shutting down message receiver...");
	stop_listening();
      }


    while(true) {
      try {
	message=linda.in(mask); //get message from Linda
	
	handle_linda_message(message);
	
      }//end try 
      catch (Exception e) {
	System.err.println("ERROR: performing 'in' on Linda");
	e.printStackTrace();
	System.err.println("Stopping..");
	stop_listening();
      }
    }//end while(true)
  }//end run
  

  /** Handle a raw string message (as long as it represents 
    * a kraft_msg term it will be fine)
    */
  //----------------------------------------------------
  public void handle_linda_message(String raw_message) {
  //----------------------------------------------------
    try{
      TermInString term = new TermInString(raw_message);

      //    if(term.functor.equals("monitor") && term.arity==2)
      //      term = new TermInString(term.argument(2));
      
      if(term.functor.equals("kraft_msg") && term.arity==2) {
	CCQL ccql_msg = new CCQL(term.argument(2));

	//Handle various types of CCQL message.
	
	if(ccql_msg.performative.equals("forward")) {
	  handle_forward(ccql_msg);
	}
	else if(ccql_msg.performative.equals("register")) {
	  handle_register(ccql_msg);
	}
	else 
	  System.err.println("Warning: unrecognised ccql performative:-\n"+
			     ccql_msg);
      }
      else System.err.println("Warning: unrecognised kraft_msg:-\n"+ 
			      term.term);
    }
    catch(Exception e) {
      System.err.println("Error handling message:-\n" + e.getMessage());
      e.printStackTrace();
    }
  }//end handle_linda_message()


  /** Handle a 'forward' predicate
    *
    */
  //----------------------------------------------------------
  protected void handle_forward(CCQL ccql_msg) throws Exception{
  //----------------------------------------------------------

    if( ! ccql_msg.performative.equals("forward"))
      throw new Exception("not a forward message");

    
    //the content of this forward is a full kraft_msg

    if(ccql_msg.language.equals("kraft_msg")) {

      System.out.println("content = " + ccql_msg.content);
      Kraft_msg km = new Kraft_msg(ccql_msg.content);
      
      if(messages.is_a_duplicate(km)==true) {
	//System.out.println("message is a duplicate -- discarded");
      }
      else{
	try{
	  System.out.println("\nReceived forwarded message:\n" +km);
	  register_agent(km.sender);
	  register_agent(km.receiver);
	  monitor.informOfMessage(km.sender, km.receiver, km);
	  System.out.println("Message monitored OK");
	}
	catch(Exception e) {
	  System.err.println("Warning: had problems Monitoring message\n"+
			     "       : error   = "+ e +":"+e.getMessage()+"\n"+
			     ": message =\n      "+ km);
	  e.printStackTrace();
	}
      }//end else
    }
    else throw new Exception("Invalid message language: '"+ccql_msg.language+"'");
  }//end handle_forward


  private void register_agent(String name) throws Exception{
    register_agent(new KraftAgent(name));
  }

  //--NB--------------------------------------------------------------
  //this could be adjusted to keep track of the agent objects by using 
  // a hashtable;
  // eg:  assume Hashtable agents; exists
  //      
  //      if( ! agents.containsKey(name)) {
  //        monitor.register(agent);
  //        agents.put(name, agent);
  //      }
  //------------------------------------------------------------------

  private void register_agent(KraftAgent agent) throws Exception{
    String name = agent.getFullName();
    
    if( ! names.contains(name)) {
      monitor.register(agent);
      names.addElement(name);
      System.out.println("Registered agent with monitor: "+name);
    }
  }



  /** handle a 'register' performative.
    *
    * Content of this message can contain info like this:
    *  [role: wrapper, resource_typ: DB, resource_protocol: sql..]
    */
  protected void handle_register(CCQL msg) {

    TermInString content;
    int type;
    KraftAgent agent;

    try{
      System.out.println("\nReceived a register msg from agent: "+msg.sender);

      agent = new KraftAgent(msg.sender);

      //System.out.println("content of msg = " + msg.content);
      content = new TermInString(msg.content);

      //We expect the content of this msg to be of a term of 
      // the form '.'( :(param,value), [] ).
      // Extract the agents ROLE from this list if possible

      //Decide on the agent type
      //-----------------------
      String role = KraftTerm.getFieldValue(content, "role");
      System.out.println("---> agent's role is: " + role);

      // an agent's resource information
      String rec_type  = KraftTerm.getFieldValue(content,"resource_type");
      String rec_proto = KraftTerm.getFieldValue(content,"resource_protocol");
      System.out.println("---> agent's resource is: " + rec_type);
      System.out.println("---> resource protocol is: " + rec_proto);

      agent.setRole(role);
      agent.setResource(rec_type, rec_proto);
     

      //Get any more parameter pairs and tell the agent about them
      String[] pair = {"w","w"};
      for(int i=1; ; i++) {
	pair = KraftTerm.getListPair(content,i);
	if(pair[0].equals("")) break;//from for loop
	if( ! (pair[0].equals("role") || 
	       pair[0].equals("resource_type") ||
	       pair[0].equals("resource_protocol"))) { //we already know about these

	  agent.addInfoPair(pair[0],pair[1]);
	}//end if
      }//end for

      if(names.contains(msg.sender)) {
	System.out.println("---> Have already registered an agent called '"+
			   msg.sender + "', ignoring it.");
      }
      else{
	//Now, register the agent with Monitor and store it.
	//--------------------------------------------------
	register_agent(agent);
      }
    }
    catch(Exception e) {
      System.err.println("Warning: could not register agent\n-->"+e);
      e.printStackTrace();
    }
  }//end handle_register


  /** Stop listening for messages and 
    * clean out any remaining duplicates intended for the Monitor
    */
  //----------------------------
  public void stop_listening() {
  //----------------------------
    
    stop();
    
    if(linda!=null) {
      try {
	cleanLinda();
	linda.disconnect();
      } catch (IOException e) { 
	System.err.println("Had probs");
	e.printStackTrace();
      }
    }//if
    System.out.println("Have stopped listening for messages");
  }//end stop_listening
  
  
  

  /** Clean up linda server -- remving extraneous messages left behind 
    *  from previous instantiations of same Monitor 
    */
  //-------------------------------------------
  private void cleanLinda() throws IOException{
  //-------------------------------------------
    
    String msg=null;
    
    //so Linda stops copying messages to me.
    linda.set_timeout(500);
    
    //eat up any remaining duplicates left for me.
    for(int i=0; i<mask.length; i++) {
      if(mask[i]!=null)
	while((msg = linda.in_noblock(mask[i])) != null) 
	  {
	    System.out.println("Cleaned old message from linda:\n--> "+msg);
	    //loop
	  }
    }
  }//end cleanLinda
  
}//end class



/** Used to keep a store of the messages currently monitored.
  * Stores them in a reduced way, in an Entry object which 
  * contains the message's sender,reciever,timestamp,reply_with,in_reply_to,
  * and the content length.
  */
class MessageHashtable extends Hashtable 
{

  MessageHashtable() {
    super(30);
  }
  
  boolean is_a_duplicate(Kraft_msg km) {

    CCQL   core = km.getCore();
    String time = km.time_stamp,
           s    = core.sender,
           r    = core.receiver,
           rw   = core.reply_with,
           irt  = core.in_reply_to;
    int    cs   = core.content.length();

    if(containsKey(time)) { // we do have this time stamp
                            //check other fields for match
      Entry entry = (Entry) get(time);
      
      if(entry.matches(s,r,rw,irt,cs))
	return true;
    }
    //if we get to here, there is no match in the table,
    // but we need to add it in for future checks.
    
    put(time, new Entry(s,r,rw,irt,cs));  //add in this one as a new
    return false;

  }//end method

    
  
  private class Entry {
    String sender, receiver, reply_with, in_reply_to;
    int    checksum;

    Entry(String s, String r, String rw, String irt, int cs){
      sender      = new String(s);  //ensure they don't get changed
      receiver    = new String(r);
      reply_with  = new String(rw);
      in_reply_to = new String(irt);
      checksum    = cs;
    }
    boolean matches(String s, String r, String rw, String irt, int cs){
      return (sender.equals(s)      && 
	      receiver.equals(r)    && 
	      reply_with.equals(rw) && 
	      in_reply_to.equals(irt) &&
	      checksum==cs);
    }
  }//end class Entry

}//end MessageHashtable
