package kraft.monitor.display;

import java.awt.*;

/** An event or ocurrence that needs to be graphically displayed 
  * by the Monitor.
  * For example; a message being passed from one agent to another is 
  * shown as an arrow, or an agent represented as circle with a name.
  *<br><br>
  * Defines methods that are important to ensure consistency is 
  * maintained when visualising these events-- and the order in which
  * they occur is also maintained.
  *<br><br>
  * Contains a DisplayMonitor -- which is used to keep track of whether 
  * an event has been displayed.
  *<br><br>
  * Used by the MonitorCanvas to help when despatching information 
  * to the relevant objects -- so all can be viewed equally.
  * 
  * @author Ted Francis
  * @see kraft.monitor.display.DisplayMonitor
  * @see kraft.monitor.display.MonitorCanvas
  * @see kraft.monitor.display.Link
  * @see kraft.monitor.display.Node
  */
//==============================
public interface DisplayEvent {
//==============================

  /** To paint the event onto the graphics context */
  public void    paint(Graphics g);

  /** To delete the event from the graphics context */
  public void    unpaint(Graphics g);

  /** Inform that the mouse button has been pressed at (x,y),
    * return true if this has an effect on this displayevent
    */
  public boolean mousePressedHasEffect(int x, int y);


 /** Inform that the mouse button has been released at (x,y),
    * return true if this has an effect on this displayevent
    */
  public boolean mouseReleasedHasEffect(int x, int y);


  /** Set the position of this event in relation to others stored by 
    * the MonitorCanvas. 
    *@see kraft.monitor.display.DisplayMonitor#setIndex
    */ 
  public void    setCurrentIndex(int ix);

  /** Set the maximum size that an object can use to display in. */
  public void    setContainerSize(Dimension d);


  /** Adds the specified display monitor to the object.
    * The DisplayEvent will then process the displayMonitor in
    * due course, and call <code> d.setDisplayed(true)</code>
    * when appropriate
    */
  public void    addDisplayMonitor(DisplayMonitor d);


  /** Instruction to the object to "tidy up" any attached display monitors
   *  and any other resources -- used as a destructor
   */
  public void    tidyUp();


  /** Get a string that is relevant to the current status of this object
    * e.g. the formatted string of a message
    */
  public String getContextString();

}

















