package kraft.monitor.display;

/**
 *  Any message object that is passed from one agent to another can be 
 *  recognised as such by making that object <b> implement </b> this
 *  interface. <br>
 *  It defines a few methods which must be implemented to ensure that
 *  Messages are treated correctly by the Monitor
 *  
 *  This interface also ensures that messages passed to the Monitor
 *  are <i> serializable </i>, so they can be passed remotely by copy
 *  using RMI.
 *
 *  The class CCQL is an example implementation of this 
 *  interface. 
 * 
 *@author Ted Francis
 *@see kraft.monitor.display.MonitorServer
 *@see kraft.monitor.adapter.CCQL
 *@see kraft.monitor.display.Monitorable
 */
//=====================================================
public interface Message extends java.io.Serializable {
//=====================================================

  /** Returns the basic command name of this message.
   *  i.e. would give 'ask-all' in a KQML ask-all performative
   */
  public String getCommandString();


  /** Returns a formatted String of the Message which gives more details
    * of the message in a 'nice' format (with tabs and newlines)
    *@param detail_level The level of detail to display the string at
    *       designers should ensure that it returns some value whatever
    *       the value of this parameter.  They could also choose to 
    *       ignore the value and always return the same string!
    */

  /*       NB: I suggest that developer allows caller to get the least 
   *           detail by specifying any negative number,
   *           and most detail by specifying a very large positive
   *    e.g.  if(detail_level <= 0) detail_level = 1; //the lowest
   *          if(detail_level > my_max) detail_level = my_max;
   */
  public String getDisplayString(int detail_level);


  /** Gets all printable details of the Message.
   */
  public String toString();


  /** Must be implemented to ensure comparison for equality with 
    * other Message objects are successful. 
    * Used when searching through logs of messages
    */
  public boolean equals(Object o);

}














