package kraft.monitor.display;

import kraft.monitor.tools.*;

import java.rmi.RemoteException;
import java.awt.event.*;
import java.awt.*;

import java.util.Date;


/** This is the backbone to the Monitor program.
  * 
  * Users must provide an object that calls the monitoring methods
  * made available via the MonitorServer interface.
  * This is especially important if the methods are intended to be called
  * remotely, as this class does not perform the necessary binding to a
  * Naming service or exporting as a remote object (as the object would 
  * be unnecessarily large, a smaller object should be used for this purpose)
  *
  * See LinjaWrapper for an example of a calling object that does
  * more -- it is not intended to be remote to this GUI as it hooks 
  * into a linda server, and receives duplicate messages, which it then 
  * converts into a format usable by this GUI (Message/Monitorable interfaces).
  *
  * Provides a graphical interface which forms a useful visualisation tool
  * to visualise the interactions in a co-operating agent system.
  * <br><br>
  * Also deals with providing a Graphical User Interface, with buttons and
  * a slider to allow stepwise movement (backwards and forwards) through
  * the visualisation
  * <br><br>
  *
  *@author Ted Francis
  *@see kraft.monitor.display.MonitorCanvas
  *@see kraft.monitor.adapter.LinjaWrapper
  */
//=====================================================================
public class MonitorGUI
  extends Panel  
  implements ActionListener ,AdjustmentListener, ItemListener, MonitorServer{
//=====================================================================
    
    /** Constants for use with the <code> displayEvent </code> method.
      *@see #displayEvent
      */
    public static final int 
      MESSAGE_SENT   = 0,
      MESSAGE_UNSENT = 1,
      STATUS_CHANGE  = 2;
    
    /** Constants for use with the <code> setDisplayMode </code> method.
      *@see #setDisplayMode
      */
    public static final int  
      STEP_BY_STEP = 1,
      TIME_STEP    = 2;
  
    /** Constants for use with the <code> setDelay </code> method.
      *@see #setDelay
      */
    public static final long
      SLOW         = 1500,
      MEDIUM       = 1000,
      FAST         = 500,
      REAL_TIME    = 0;
    
    protected MonitorCanvas      canvas;
    
    private static int  FONT_SIZE = 12;
    private int    displayMode    = STEP_BY_STEP;
    private long   DELAY          = MEDIUM;    //default delay between showing 
    private int    LAST           = -911;      // steps is 1sec 
    
    
    //Components of the GUI
    //------------------------
    private TextArea           top_text, lower_text;
    private Scrollbar          slider;
    private Label              sliderLabel, modeLabel, speedLabel;
    private GridBagLayout      layout;
    private GridBagConstraints gbc;
    
    private Button             clearTextButton, pauseButton;
    private Choice             modeChoice, speedChoice, optionsChoice;
    
    
    public MonitorGUI() {
      init();
    }
    
    /** Initialise the GUI, the MonitorCanvas */ 
    //-------------------
    public void init(){
    //-------------------
      
      // Set the layout for the GUI
      // -----------------------------
      
      layout = new GridBagLayout();
      gbc    = new GridBagConstraints();
      this.setLayout(layout);
      this.setFont(new Font("SanSerif", Font.PLAIN, FONT_SIZE));
      
      
      // Set up the components
      // ---------------------
      
      //the text areas shows text logging of messages received.
      top_text = new TextArea("", 2 , 80 , TextArea.SCROLLBARS_VERTICAL_ONLY);
      top_text.setEditable(false);
      top_text.setVisible(true);
      top_text.setForeground(Color.blue);
      top_text.setBackground(Color.white);
      
      
      //the text area shows text logging of messages received.
      lower_text = new TextArea("", 7 , 80 , TextArea.SCROLLBARS_VERTICAL_ONLY);
      lower_text.setEditable(false);
      lower_text.setVisible(true);
      lower_text.setForeground(Color.black);
      lower_text.setBackground(Color.white);
      
      //Button to go to previous step in visualisation
      Button previousButton = new Button("< Prev");
      previousButton.setActionCommand("prev");
      previousButton.addActionListener(this);
      
      //Button to go to next step in visualisation
      Button nextButton = new Button("Next >");
      nextButton.setActionCommand("next");
      nextButton.addActionListener(this);
      
      // button to go to step by step mode
      pauseButton = new Button("Pause ||");
      pauseButton.setActionCommand("pause");
      pauseButton.addActionListener(this);
      pauseButton.setEnabled(false);
      
      //set all these buttons to be in bold font
      Font f = new Font(getFont().getName(), Font.BOLD, getFont().getSize());
      previousButton.setFont(f);
      nextButton.setFont(f);
      pauseButton.setFont(f);
      
      
      // Panel containing all the above buttons
      Panel navPanel = new Panel(new FlowLayout(FlowLayout.RIGHT));
      navPanel.add(previousButton);
      navPanel.add(pauseButton);
      navPanel.add(nextButton);
      
      // Set up a pull-down choice box for options
      optionsChoice = new Choice();
      optionsChoice.addItemListener(this);
      optionsChoice.add("Options...");
      optionsChoice.add("Hide Text");
      optionsChoice.add("Clear Text");
      optionsChoice.add("Clear Messages");
      
      Panel optionsPanel = new Panel(new FlowLayout(FlowLayout.RIGHT));
      optionsPanel.add(optionsChoice);
      
      
      // Initialise the MonitorCanvas, ready to display
      canvas = new MonitorCanvas();
      canvas.setVisible(true);
      
      //Here is a slider to display the current visual step we are at
      //-------------------------------------------------------------
      slider = new Scrollbar(Scrollbar.HORIZONTAL);
      slider.setValues(0, 5, 0, 5);
      slider.setUnitIncrement(1);
      slider.setBlockIncrement(3);
      slider.setEnabled(true);
      slider.addAdjustmentListener(this);
      
      
      //Some labels to show current status
      sliderLabel = new Label("Step: 0 of 0",                    Label.LEFT);
      modeLabel   = new Label("Mode: "+stringValue(displayMode), Label.CENTER);
      speedLabel  = new Label("Speed: "+stringValue(DELAY),      Label.RIGHT);
      speedLabel.setEnabled(false);
      sliderLabel.setFont(new Font(getFont().getName(),
				   Font.BOLD,
				   getFont().getSize()));
      
      
      
      //A pull-down choice box to switch between display modes.
      modeChoice = new Choice();
      modeChoice.addItemListener(this);
      modeChoice.add(stringValue(STEP_BY_STEP));
      modeChoice.add(stringValue(TIME_STEP));
      modeChoice.select(stringValue(displayMode));
      
      Panel modePanel = new Panel(new FlowLayout(FlowLayout.CENTER));
      modePanel.add(new Label("Mode:", Label.RIGHT));
      modePanel.add(modeChoice);
      
      
      //A pull-down choice box to switch between display speeds.
      speedChoice = new Choice();
      speedChoice.addItemListener(this);
      speedChoice.add(stringValue(SLOW));
      speedChoice.add(stringValue(MEDIUM));
      speedChoice.add(stringValue(FAST));
      speedChoice.add(stringValue(REAL_TIME));
      speedChoice.select(stringValue(DELAY));
      speedChoice.setEnabled(false);
      
      Panel speedPanel = new Panel(new FlowLayout(FlowLayout.RIGHT));
      speedPanel.add(new Label("Speed:",Label.RIGHT));
      speedPanel.add(speedChoice);
      
      //OK, start putting in the components to this, giving the different
      // grid x,y,width,heights needed.
      //-----------------------------------------------------------------
      
      gbc.fill   = GridBagConstraints.BOTH;
      gbc.weightx = 1.0;
      gbc.weighty = 1.0;
      
      add(canvas,      0,1, GridBagConstraints.REMAINDER, 7);
      
      gbc.weightx = 0.5;
      gbc.weighty = 0.0;
      
      add(sliderLabel,  0,8, 1, 1);
      add(modePanel,   1,8, 1, 1);
      add(speedPanel,  2,8, 1, 1);
      
      gbc.weightx = 0.0;
      add(slider,       0, 9, GridBagConstraints.REMAINDER, 1);
      add(navPanel,     0,10, 2, 1);
      add(optionsPanel, 2,10, 1, 1);
      add(top_text,     0,11, GridBagConstraints.REMAINDER, 1);
      add(lower_text,   0,12, GridBagConstraints.REMAINDER, 1);
      
      
      doLayout();
      
    }//end init
    
    
    public void stop() {
      
      //do anything in here that should be de-constructed,
      // freed-up etc..
      
    }
      

    /** The the set of Messages currently being shown by 
      * the Monitor
      */
    public String[] getMessages() {
      return canvas.getMessages();
      
    }

    
    
    /** Add the component to this GUI using x,y,w and h to specify
      * their gridx, gridy etc in the GridBagConstraints
      */  
    //------------------------------------------------------------
    private void add(Component comp, int x, int y, int w, int h) {
    //------------------------------------------------------------
      
      gbc.anchor  = GridBagConstraints.CENTER;
      gbc.gridx   = x;
      gbc.gridy   = y;
      gbc.gridwidth = w;
      gbc.gridheight= h;
      
      layout.setConstraints(comp, gbc);
      this.add(comp);
    }
    
    /** Adjust the slider according to the current step and total steps
      * available in the MonitorCanvas 
      */
    //--------------------------
    private void adjustSlider(){
    //--------------------------
      
      int val, vis, min, max;
      
      val = canvas.currentStep();
      min = 0;
      max = canvas.totalSteps();
      vis = 1;
      
      
      slider.setValues(val, vis, min, max);
      
      if(max<=0) 
	sliderLabel.setText("Step: 0 of 0");
      else
	sliderLabel.setText("Step: " + (val+1) + " of " + max);
      
      updateText();
      /*Debug.println("-------adjustSlider()----------------");
	Debug.println("Slider is now: [val = "+val+", vis = "+vis+
	", min = "+min+", max = "+max);
	Debug.println("Label is now: "+sliderLabel.getText());
      */
      
    }//end adjustSlider
    
    
    /** Output the string to the top text area shown by this GUI 
      */
    public void output(String string){
      top_text.setText(string);
      lower_text.setText("");
    }


    /**Constant used in output()
      *@see #output(int)
      */
    public static final int TOP=0, BOTTOM=1;


    /** Output the string in one of the text areas shown by this GUI
      *@param where either TOP or BOTTOM
      */
    public void output(int where, String string) {
      switch(where) {
      case TOP: top_text.setText(string); break;
      case BOTTOM: lower_text.setText(string); break;
      }
    }
	
    
    /** Set whether the GUI should show each visualisation step
      * and wait for the user to click 'next' - or whether it should 
      * run through the steps according to a particular delay 
      *@see #STEP_BY_STEP
      *@see #TIME_STEP
      */
    //------------------------------------
    public void setDisplayMode(int mode) {
    //------------------------------------
      displayMode = mode;
      modeLabel.setText("Mode: "+ stringValue(mode));
      modeChoice.select(stringValue(mode));
      
      switch(displayMode) {
      case STEP_BY_STEP: 
	speedLabel.setEnabled(false);
	speedChoice.setEnabled(false);
	pauseButton.setEnabled(false);
	pauseButton.setActionCommand("cont");
	break;
      case TIME_STEP:
	speedLabel.setEnabled(true);
	speedChoice.setEnabled(true);
	pauseButton.setEnabled(true);
	pauseButton.setActionCommand("pause");
	gotoStep(LAST);
	adjustSlider();
	break;
      }
    }
  
    /** Set the time delay for use in TIME_STEP mode.
      *@see #SLOW
      *@see #MEDIUM
      *@see #FAST
      *@see #REAL_TIME
      */
    //---------------------------------
    public void setDelay(long millis) {
    //---------------------------------
      DELAY = millis;
      speedLabel.setText("Speed: "+ stringValue(millis));
    }
    
 
    private String current=null;
   

    /** update the text in the windows according to the 
      * canvas' current context string
      */
    private void updateText() {
      String str = 
	canvas.getContextString();  //whatever the canvas wants to display
      
      if(str==null) {
	output(TOP,"");
	output(BOTTOM,"");
	return;
      }

      if( ! str.equals(current)) {
	int ix = str.indexOf("\n\n");
	if(ix < 0) {
	  output(TOP,"");
	  output(BOTTOM,str);
	}
	else {
	  output(TOP,   str.substring(0,ix));
	output(BOTTOM,str.substring(ix+2,str.length()));
	}
      }
      current = str;
    }
    
    
    private void nextStep() {
      canvas.nextStep();
    }
    
    private void previousStep() {
    canvas.previousStep();
    }
    
    private void gotoStep(int val){
      if(val==LAST) 
	canvas.gotoStep(canvas.totalSteps()-1);
      else
	canvas.gotoStep(val);
      
    }
    
    private void clearMessages() {
      canvas.clearEvents();
    }
    
    /** Give a string to represent one of the constants in capitals */
    private String stringValue(int num){
      switch(num){
	//values for display modes
      case STEP_BY_STEP:   return "Step by step";
      case TIME_STEP:      return "Continuous  ";
	
	//values for speeds
      case (int)SLOW:      return "Slow     ";
      case (int)MEDIUM:    return "Medium   ";
      case (int)FAST:      return "Fast     ";
      case (int)REAL_TIME: return "Real Time";
	
	//default value
      default:             return "unknown";
      }
    }

    private String stringValue(long num){
      return stringValue((int)num);
    }//end speedStr
    
    
    /** Recieves AdjustmentEvents related to the slider used to step 
      * between visualisation steps.  Handles the events correctly.
      */
    //-------------------------------------------------------
    public void adjustmentValueChanged(AdjustmentEvent evt) {
    //-------------------------------------------------------
      
      Adjustable adj = evt.getAdjustable();
      
      if (adj == slider) {   //it MUST be as there are no other Adjustable
	
	gotoStep(evt.getValue());
	adjustSlider();
	
      }//end if(adj==slider)
      
    }//end adjustmentValueChanged
    
    
    /** Receives ActionEvents from the buttons in the GUI and handles them*/
    //-----------------------------------------
    public void actionPerformed(ActionEvent e){
    //-----------------------------------------
      
      String command = e.getActionCommand();
      
      if(command.equals("pause")){
	Debug.println("Step by step mode selected");
	setDisplayMode(STEP_BY_STEP);
      }
      else if (command.equals("cont")){
	Debug.println("Continuous mode selected");
	setDisplayMode(TIME_STEP);
      }
      else if (command.equals("next")){
	Debug.println("'Next' button selected");
	nextStep();
	adjustSlider();
      }
      else if (command.equals("prev")) {
	Debug.println("'Previous' button selected");
	previousStep();
	adjustSlider();
      }
      
    }//end actionPerformed

    
  /** Recieves ItemEvents from the choice boxes in the GUI, and
    * handles them accordingly
    */
    //-------------------------------------------
    public void itemStateChanged(ItemEvent evt) {
    //-------------------------------------------
      String         command;
      ItemSelectable item = evt.getItemSelectable();
      
      if(item == modeChoice){
	command = modeChoice.getSelectedItem();
	
	if(command.equals(stringValue(STEP_BY_STEP))){
	  Debug.println("Step by step mode selected");
	  setDisplayMode(STEP_BY_STEP);
	  speedChoice.setEnabled(false);
	}
	else if (command.equals(stringValue(TIME_STEP))){
	  Debug.println("Continuous mode selected");
	  setDisplayMode(TIME_STEP);
	  speedChoice.setEnabled(true);
	}
      }
      else if(item == speedChoice){
	command = speedChoice.getSelectedItem();
	if (command.equals(stringValue(SLOW))){
	  Debug.println("Slow selected, delay = "+SLOW);
	  setDelay(SLOW);
	}
	else if (command.equals(stringValue(MEDIUM))){
	  Debug.println("Medium selected, delay = " + MEDIUM);
	  setDelay(MEDIUM);
	}
	else if (command.equals(stringValue(FAST))){
	  Debug.println("Fast selected, delay = " + FAST);
	  setDelay(FAST);
	}
	else if (command.equals(stringValue(REAL_TIME))){
	  Debug.println("Real Time speed selected");
	  setDelay(REAL_TIME);
	}
      }
      else if(item == optionsChoice) {
	command = optionsChoice.getSelectedItem();
	int sel = optionsChoice.getSelectedIndex();
	if(command.equals("Hide Text")) {
	  top_text.setVisible(false);
	  lower_text.setVisible(false);
	  optionsChoice.remove(sel);
	  optionsChoice.insert("Show Text", sel);
	  doLayout();
	}
	else if(command.equals("Show Text")) {
	  top_text.setVisible(true);
	  lower_text.setVisible(true);
	  optionsChoice.remove(sel);
	  optionsChoice.insert("Hide Text", sel);
	  doLayout();
	}
	else if(command.equals("Clear Text")) {
	  top_text.setText("");
	  lower_text.setText("");
	}
	else if (command.equals("Clear Messages")){
	  clearMessages();
	  adjustSlider();
	}
	optionsChoice.select(0);
      }
      
    }//end itemStateChanged
    
    
    //-------------------------------------------------------------------------  
    /** Informs the monitor of an 'event' to be displayed.
      * This is synchronized to enable unique time-stamping of these events, 
      * so as to ensure consistency when tracing back/forwards through 
      * the currently stored events.
      * The type of event specified in <code>event_type</code> dictates the 
      * content of the other parameters. The appropriate method is called
      * to display the various types of events
      *<br><br>
      * Depending on the current display mode and time delay in the Monitor, this
      * method will only return when this 'event' has been displayed.  This is to ensure
      * that what is seen on the display is what is currently  occuring in the agent 
      * scenario -- as agents are 'blocked' until the event they wish to visualise has
      * has been displayed. 
      *
      *@param event_type The type of event, valid values are <code> MESSAGE_SENT,
      *                  MESSAGE_UNSENT, STATUS_CHANGE </code>.
      *@param from       The name of the Monitorable object that this event 
      *                  concerns   
      *@param otherStr   In the case of <code>STATUS_CHANGE</code>, this is the 
      *                  latest comment string sent by 'from'.
      *                  In the other cases, this is the recipient of the Message
      *                  specified in 'msg'.
      *@param msg        In the case of <code>STATUS_CHANGE</code>, this is 
      *                  ignored.  In the other cases, this is the Message object
      *                  passed from 'from' to 'otherStr'
      *
      *@see kraft.monitor.display.Monitor#MESSAGE_SENT
      *@see kraft.monitor.display.Monitor#MESSAGE_UNSENT
      *@see kraft.monitor.display.Monitor#STATUS_CHANGE
      *@see kraft.monitor.display.Monitor#displayMessage
      *@see kraft.monitor.display.Monitor#crossOutMessage
      *@see kraft.monitor.display.Monitor#statusChange
      */
    //-------------------------------------------------------------------  
    public synchronized void displayEvent(int  event_type, String from, 
					  String otherStr, Message msg) {
    //-------------------------------------------------------------------
    
      switch(event_type) {
      case MESSAGE_SENT:    displayMessage(from, otherStr, msg); break;
      case MESSAGE_UNSENT: crossOutMessage(from, otherStr, msg); break;
      case STATUS_CHANGE:     statusChange(from, otherStr);      break;
      }
    }//end displayEvent
    
    
    
    
    /** Indicates a change in status of a registered Monitorable object.
      * Notifies the MonitorCanvas of the change.
      *@param name The name of the Monitorable object the change affects.
      *@param comment The detail of the status change
      *@see   kraft.monitor.display.MonitorCanvas#statusChange
      */
    //--------------------------------------------------------
    protected void statusChange(String name, String comment) {
    //--------------------------------------------------------
      Date date = new Date(); //this gets the current date/time
      
      DisplayMonitor currentDisplayMonitor;
      
      output("** At " + date + ": status update for \"" +
	     name + "\" received:\n\t" + comment);
      
      /* Set a displayMonitor, keep a handle to it so we are notified 
	 when this Change has been displayed. */
      
      currentDisplayMonitor = new DisplayMonitor(date);
      
      canvas.statusChange(name, comment, currentDisplayMonitor);
      adjustSlider();
      
      
      switch(displayMode) {
	
      case STEP_BY_STEP:	
	while ( currentDisplayMonitor.isDisplayed() == false ) { 
	  //sleep until the message gets displayed
	  try{
	    Thread.sleep(800);
	  }
	  catch(InterruptedException e) {};
	  if(displayMode==TIME_STEP) //changed by user clicking button
	    break; //from while loop
	}//end while
	break;
	
	
      case TIME_STEP:	  
	try{
	  Thread.sleep(DELAY);
	}catch(InterruptedException e){}
	gotoStep(currentDisplayMonitor.getIndex());
	adjustSlider();
	break;
	
    }//end switch
      
    }
    
    
    
    /** Used to indicate that some failure has ocurred when delivering a 
      * Message.
      * Notifies the MonitorCanvas of this.
      * A Message that has previously been displayed
      * did not reach its destination, the effect of calling this will be 
      * to 'cross out' the graphical representation of the message if it is 
      * displayed  on the MonitorCanvas again.
      *
      *@see kraft.monitor.display.MonitorCanvas#crossOutMessage
      */
    
    //----------------------------------------------------------------
    protected void crossOutMessage(String from, String to, Message msg) {
    //----------------------------------------------------------------
      canvas.crossOutMessage(from, to, msg);
    }
    
    
    
    /** Display a copy of a Message sent by a monitorable object.
      * The names 'from' and 'to' are checked against registered Monitorables
      * and the 'event' is queued for displaying.
      * The MonitorCanvas is notified of this Message, and displays it 
      * according to the current display mode and time delay.
      * To ensure consistency with timing and race conditions when displaying
      * different events, this method is protected. Use 
      * <code> displayEvent(MonitorGUI.MESSAGE_SENT, from, to, msg) </code>
      * from non subclasses of this class.
      *
      *@see kraft.monitor.display.MonitorGUI#displayEvent
      *@see kraft.monitor.display.MonitorGUI#MESSAGE_SENT
      *@see kraft.monitor.display.MonitorGUI#setDisplayMode
      *@see kraft.monitor.display.MonitorGUI#setDelay
      */
    //-----------------------------------------------------------------  
    protected void displayMessage(String from, String to, Message msg){
    //-----------------------------------------------------------------  
      
      Date date = new Date(); //this gets the current date/time
      
      DisplayMonitor currentDisplayMonitor;
      
      
      /* Set a displayMonitor, keep a handle to it so we are notified 
	 when this Message has been displayed. */
      
      currentDisplayMonitor = new DisplayMonitor(date);
      
      canvas.displayMessage(from, to, msg, date, currentDisplayMonitor);
      adjustSlider();
      
      
      switch(displayMode) {
	
      case STEP_BY_STEP:	
	while ( currentDisplayMonitor.isDisplayed() == false ) { 
	  //sleep until the message gets displayed
	  try{
	    Thread.sleep(800);
	  }
	  catch(InterruptedException e) {};
	  if(displayMode==TIME_STEP) //changed by user clicking button
	    break; //from while loop
	}//end while
	break;
	
	
      case TIME_STEP:	  
	try{
	  Thread.sleep(DELAY);
	}catch(InterruptedException e){}
	gotoStep(currentDisplayMonitor.getIndex());
	adjustSlider();
	break;
	
      }//end switch
    }//end displayMessage
    
    
    /* These are the methods needed to implement the MonitorServer
     * interface.
     * Provides common interface for 'wrapper' objects (eg. LinjaWrapper)
     * to call methods on the Monitor.
     */
    

    /** Registers a monitorable object that wishes to be monitored 
      * @see MonitorServer 
      */
    //------------------------------------------------------------
    public void register(Monitorable agent) throws RemoteException {
    //------------------------------------------------------------
      String name = "unknown_name";
      int    type = Monitorable.MIDDLE;
      
      Debug.println("MonitorGUI.register: entered");
      try {
	name = agent.getFullName();
	type = agent.getType();
	output(TOP,    "Have registered an agent at " + Debug.time());
	output(BOTTOM, agent.getDisplayName(2));
      }
      catch(RemoteException e) {
	System.err.println("Could not find name/type of Monitorable on register");
	//throw e;
      }
      canvas.add(name, agent, type);
      //---------------
      // We only need to adjust the slider if an agent de-registering
      // is considered as a DisplayEvent -- it is not, so we don't!
      //
      //adjustSlider();
      //---------------
    }//end register
    
    
    /** Make the monitorable object with this name unmonitorable 
      * @see MonitorServer 
      */
    //----------------------------------
    public void deRegister(String name) {
    //----------------------------------
      canvas.remove(name);
      output(TOP,"An agent has unregistered at "+Debug.time());
      output(BOTTOM, "name:  " + name);
      //---------------
      // We only need to adjust the slider if an agent de-registering
      // is considered as a DisplayEvent -- it is not, so we don't!
      //
      //adjustSlider();
      //---------------
      Debug.println("Have deregistered an agent: "+name);
    }
    

    /** @see MonitorServer */
    public void informOfMessage(String from, String to, Message m) 
    {
      displayEvent(MonitorGUI.MESSAGE_SENT, from, to, m);
    }
    
    
    /** @see MonitorServer */
    public void informOfFailure(String from, String to, Message m)
    {
      
      displayEvent(MonitorGUI.MESSAGE_UNSENT, from, to, m);
    }
    
    
    /** @see MonitorServer */
    public void informOfStatus(String name, String comment) 
    {
      
      displayEvent(MonitorGUI.STATUS_CHANGE, 
		   name, comment, null);
    }

}//end MonitorGUI


















