/* Copyright (C) 1990 Swedish Institute of Computer Science */

/*

	Modified to use the Aberdeen socket library
	for SICStus & Quintus Prolog

	Modified by:	K.Y. Hui
	Version:	1.02
	Date:		12 November 1997
*/


:- module(linda_client,[
	linda_client/1,
	linda_client/2,
	close_client/0,
	in/1,
	in/2,
	in_noblock/1,
	in_noblock/2,
	out/1,
	rd/1,
	rd/2,
	rd_noblock/1,
	rd_noblock/2,
        bagof_rd_noblock/3,
	linda_timeout/2
   ]).

:- prolog_flag(character_escapes,_,on).		% allow '\n' notation

:- use_module('util/prolog_detect').

:- prolog_detect:prolog(Prolog),
	(	(Prolog=sicstus,
		!,
		use_module(library(lists))
		);
		(Prolog=quintus,
		use_module(library(basics))
		)
	).


:- use_module('socket_lib/sockets').
:- use_module('fastrw_lib/fastrw').

:- dynamic linda_stream/4.
:- dynamic time_out/1.
:- assert(time_out(20:0)).

%protocol(0'p).
protocol(0'f).      /* 0'f = special format, 0'p = write_canonical */

linda_client(Host,Port) :-
	linda_client(Host:Port).

linda_client(Address) :-
	linda_stream(Address,_,_,_),
	!.

linda_client(Address) :-
	linda_stream(_,_,_,_),
	!,
	format(user_error,"ERROR: linda_client/1: already client to ~w~n",[Address]).

linda_client(Address) :-
	open_client(Address,Socket,R,W),
	assertz(linda_stream(Address,Socket,R,W)),
	ping(Answer),
	ping_answer_ok(Answer).


%
% check if ping is ok
%
ping_answer_ok(pong) :- !.

ping_answer_ok(A) :-
	format(user_error,
	       '{ERROR: linda_client/1: strange answer from server: ~q}~n',
	       [A]).	



%
% open connection to Linda server
%
open_client(Host:Port,Socket,R,W) :-
	atom(Host),
	integer(Port),
	!,
	socket_create('AF_INET',Socket),			%create socket
	socket_connect(Socket,'AF_INET'(Host,Port),R,W).	%connect to remote

%
% fail to connect
%
open_client(Addr,_,_) :-
	format(user_error,
	       '{ERROR: open_client/3: Illegal network address: ~w}~n',
	       [Addr]),
	fail.


%-----------------------------------------------------------------------------
close_client :-
	set_input(user_input),
	set_output(user_output),
	retract(linda_stream(_,Socket,_,_)),
	socket_destroy(Socket).					%close socket


ping(Answer) :-	
	to_linda(0'p, ping),
	time_out_select,
	from_linda(Answer).

out(T) :-
	to_linda(0'o, T).

%
% "in" that waits forever
%
in(T) :-
	repeat,			%loop when backtrack
        to_linda(0'i, T),
	from_linda(T).

%
% "in" with a list of templates
%
in(Tuples,Tuple) :-
	Tuples = [_|_],
	repeat,			%loop when backtrack
        to_linda(0'I, Tuples),
	from_linda(Tuple),
	member(Tuple,Tuples).  % for unification of Tuples with answer


%
% "in" without blocking
%
in_noblock(T) :-
	to_linda(0'j, T),
	time_out_select,
	from_linda(0's, T).


%
% in_noblock w/ timeout
%
in_noblock(T,Timeout) :-
	to_linda(0'i, T),		%like an "in"
	time_out_select(Timeout),	%but something must return within "timeout"
	!,
	from_linda(T).			%successfully get result within timeout

in_noblock(_,_) :-
	linda_stream(Address,_,_,_),	%if no result is got within timeout
	close_client,			%destroy this client
	linda_client(Address),		%and create a new one
	!,
	fail.				%make it fail

%
% "rd" without blocking
%
rd_noblock(T) :-
	to_linda(0's, T),
	time_out_select,
	from_linda(0's, T).


%
% rd_noblock/2 w/ timeout
%

rd_noblock(T,Timeout) :-
	to_linda(0'r, T),		%still a "rd"
	time_out_select(Timeout),	%but something must return within timeout
	!,
	from_linda(T).

rd_noblock(_,_) :-
	linda_stream(Address,_,_,_),	%if no result is got within timeout
	close_client,			%close this client
	linda_client(Address),		%create a new one
	!,
	fail.				%force it to fail


rd(T) :-
	to_linda(0'r, T),
	from_linda(T).

rd(Tuples,Tuple) :-
	Tuples = [_|_],
        to_linda(0'R, Tuples),
	from_linda(Tuple), !,
	member(Tuple,Tuples).  % for unification of Tuples with answer

bagof_rd_noblock(Template,Tuple,Bag) :- 
	to_linda(0'b, b(Template,Tuple,Bag)),
	time_out_select,
	from_linda(Bag).

linda_trace(OnOff) :-
	(var(OnOff) ; OnOff=on ; OnOff=off), !,
	to_linda(0't, OnOff),
	time_out_select,
	from_linda(OnOff).

linda_call(Goal) :-
	to_linda(0'c, Goal),
	from_linda(0's, Goal).

%-----------------------------------------------------------------------------

to_linda(Code) :-
	linda_stream(_,_,R,W),				%get the streams
	current_input(Current_in),			%get current output stream
	current_output(Current_out),			%get current output stream
	set_input(R),					%set I/O to socket streams
	set_output(W),
	protocol(P),					%get protocol to use
	write_out(P,Code),				%output data
	flush_output(W),				%flush output
	set_input(Current_in),
	set_output(Current_out),
	!.


to_linda(Code,Item) :-
	linda_stream(_,_,R,W),				%get the streams
	current_input(Current_in),			%get current output stream
	current_output(Current_out),			%get current output stream
	set_input(R),					%set I/O to socket streams
	set_output(W),
	protocol(P),					%get protocol to use
	write_out(P,Code,Item),				%output data
	flush_output(W),				%flush output
	set_input(Current_in),
	set_output(Current_out),
	!.


%-----------------------------------------------------------------------------

write_out(Protocol,Code) :- 
	member(Protocol,[0'p,0'f]),
	put(Code).

write_out(0'p,Code,Item) :- 
	put(0'p),
	put(Code), 
	write_canonical(Item),
	write('.\n').		%use '\n' instead of nl/0 because of incompatibility
				%problem in different platforms

write_out(0'f,Code,Item) :-
	put(0'f), 
	put(Code), 
	fast_write(Item).

read_in(0'p,Item) :- read(Item).
read_in(0'f,Item) :- fast_read(Item).
%-----------------------------------------------------------------------------

%
% read item from Linda
%

from_linda(Item) :-
	linda_stream(_,_,R,W),
	current_input(Current_in),			%save current I/O streams
	current_output(Current_out),
	set_input(R),					%connect to socket streams
	set_output(W),
	protocol(P),					%get protocol to use
	read_in(P,Item),				%read data item
	set_input(Current_in),				%switch back to original streams
	set_output(Current_out),
	!.

from_linda(Code,Item) :-
	linda_stream(_,_,R,W),
	current_input(Current_in),			%save current I/O streams
	current_output(Current_out),
	set_input(R),					%connect to socket streams
	set_output(W),
	get0(Cl),
	(	(Cl=Code,
		protocol(P),
		read_in(P,Item),
		set_input(Current_in),
		set_output(Current_out)
		);
		(set_input(Current_in),
		set_output(Current_out),
		fail
		)
	),
	!.



%-----------------------------------------------------------------------------
time_out_select :-
	time_out(TimeOut),
	time_out_select1(TimeOut).

time_out_select(Timeout) :-
	time_out_select1(Timeout).

time_out_select1(off) :- !.

time_out_select1(TimeOut) :-
	TimeOut = Sec:Usec,
	linda_stream(_,Socket,_,_),
	socket_select([Socket],[],[],time(Sec,Usec),[Socket],_,_),
	!.
/*
time_out_select1(_:_) :-
	format(user_error,'{ERROR: wait for linda timed out}~n',[]),
	fail.
*/

%-----------------------------------------------------------------------------

linda_timeout(Old, New) :-
	linda_timeout_arg(Old, New),
	time_out(Old),
	retract(time_out(Old)),
	assert(time_out(New)).

linda_timeout_arg(O, N) :- var(N), !, O==N.
linda_timeout_arg(_, S:M) :- integer(S), integer(M), S>=0, M>=0, !.
linda_timeout_arg(_, off).
