/*
socket library for SICSTUS Prolog

This program borrows much idea from the original Sicstus socket library.
Any error is reported by the SP_syserror function which raises an exception
to Prolog.

By: K.Y. Hui
Date: 21 Aug 1996
*/


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <netdb.h>
#include <string.h>
#include <sicstus/sicstus.h>

/*return success or failure*/

#define	SUCCESS	0
#define	FAILURE	1


/*Prolog stream record for socket*/

typedef struct	{
		int sock;
		FILE *r_stream;
		FILE *w_stream;
		} socket_rec;

/*
-------------------
function prototypes
-------------------
*/

int c_socket_create_I(long int *sock);
int c_socket_create_U(long int *sock);
int c_socket_destroy(long int sock);
int c_socket_bind_I(long int sock,long int port,long int *used_port);
int c_socket_bind_U(long int sock,char *path);
int c_socket_listen(long int sock,long int length);
int c_socket_accept(long int sock,long int *income,SP_stream **stream);
SP_stream *new_socket_stream(int sock,char *socket_name);
int c_socket_connect_I(long int sock,char *host,long int port,SP_stream **stream);
int c_socket_connect_U(long int sock,char *path,SP_stream **stream);
int c_socket_select(SP_term_ref rs,SP_term_ref ws,SP_term_ref es,long int sec,long int usec,SP_term_ref rrs,SP_term_ref rws,SP_term_ref res);
void set_fd_set(SP_term_ref list,fd_set *fset,int *max_fd);
void decode_fd_set(int max_fd,fd_set *fset,SP_term_ref result_list);
int c_hostname(char **host);

/*
I/O functions for SICSTUS stream
*/
int socket_fgetc();
int socket_fputc();
int socket_flush();
int socket_eof();
void socket_clrerr();
int socket_close();


char buffer[128];		/*buffer for returning string to Prolog*/

/*
--------------------------------------
create a socket in the Internet domain

sock=pointer to socket to be returned
--------------------------------------
*/

int c_socket_create_I(long int *sock)
{
int s;

if ((s=socket(AF_INET,SOCK_STREAM,0))<0)
	return(FAILURE);
*sock=s;
return(SUCCESS);
}


/*
** create a socket in the UNIX domain
*/

int c_socket_create_U(long int *sock)
{
int s;

if ((s=socket(AF_UNIX,SOCK_STREAM,0))<0)
	return(FAILURE);
*sock=s;
return(SUCCESS);
}



/*
---------------------------
destroy a socket
sock=socket to be destroyed
---------------------------
*/

int c_socket_destroy(long int sock)
{
if (close(sock)!=0)
	return(FAILURE);
return(SUCCESS);
}



/*
----------------------------------------------------------------
binds a socket to an address in the Internet domain
sock=		socket to be binded
port=		port no.
		If it is 0, then system is free to choose a port
used_port=	actual port used
----------------------------------------------------------------
*/

int c_socket_bind_I(long int sock,long int port,long int *used_port)
{
struct sockaddr_in name;
int length;

name.sin_family=AF_INET;		/*internet domain*/
name.sin_addr.s_addr=INADDR_ANY;	/*any incoming address*/
name.sin_port=htons(port);		/*port no.*/


/*bind socket to name*/

if (bind(sock,(struct sockaddr *)&name,sizeof(name)))
	return(FAILURE);

/*get port no. actually used*/

length=sizeof(name);
if (getsockname(sock,(struct sockaddr *)&name,&length))
	return(FAILURE);
*used_port=ntohs(name.sin_port);		/*get port no.actually used*/
return(SUCCESS);
}




/*
-------------------------------------------
binds a socket to a path in the UNIX domain

sock=	socket to be binded
path=	pathname of file for binding
-------------------------------------------
*/

int c_socket_bind_U(long int sock,char *path)
{
struct sockaddr_un name;


name.sun_family=AF_UNIX;				/*unix domain*/
if (strlen(path)>sizeof(name.sun_path)-1)		/*check path length*/
	return(FAILURE);

strcpy(name.sun_path,path);				/*copy path into structure*/
unlink(path);						/*delete file 1st*/
if (bind(sock,(struct sockaddr *)&name,sizeof(name)))	/*bind socket*/
	return(FAILURE);

return(SUCCESS);
}




/*
---------------------------------------
listen to a socket and set queue length

sock=	socket to listen
length=	length of queue
---------------------------------------
*/
int c_socket_listen(long int sock,long int length)
{
if (listen(sock,length))
	return(FAILURE);
return(SUCCESS);
}



/*
-------------------------------------------------------
accept and incoming connection

sock=	socket to accept connection from
income=	(pointer to) incoming connection to be returned
stream=	(pointer to) SICSTUS stream to be returned

*** Remark:	socket stream is bi-directional
-------------------------------------------------------
*/
int c_socket_accept(long int sock,long int *income,SP_stream **stream)
{
int income_sock;
char socket_name[20];
struct sockaddr dummy;
int dummy_length;

dummy_length=sizeof(dummy);
if ((income_sock=accept((int)sock,&dummy,&dummy_length))<0)		/*accept incoming connection*/
	return(FAILURE);

sprintf(socket_name,"socket %d",income_sock);
*stream=new_socket_stream(income_sock,socket_name);			/*open streams on incoming socket*/
*income=income_sock;
return(SUCCESS);
}


/*
** open read & write stream to a socket
** returns a single SP_stream pointer
**
** sock=	socket to open streams
** socket_name=	socket-name (seems to be needed. Get from original SICSTUS socket codes)
*/

SP_stream *new_socket_stream(int sock,char *socket_name)
{
FILE *r,*w;
SP_stream *stream;
socket_rec *sp;

r=(FILE *)fdopen(sock,"r");		/*open read stream to socket*/
w=(FILE *)fdopen(sock,"w");		/*open write stream to socket*/
if (r==NULL || w==NULL)
	return(NULL);

setbuf(r,NULL);
sp=(socket_rec *)SP_malloc(sizeof(socket_rec));	/*allocate memory for structure*/
sp->r_stream=r;
sp->w_stream=w;
sp->sock=sock;					/*put socket into structure*/

/*install prolog stream*/

SP_make_stream(	sp,
		socket_fgetc,
		socket_fputc,	
		socket_flush,
		socket_eof,
		socket_clrerr,
		socket_close,
		&stream);
stream->filename=socket_name;
stream->modename="socket";
stream->fd=sock;
return(stream);
}


/*
** set of I/O functions for socket stream
*/

/*
** read next char from stream
*/
int socket_fgetc(socket_rec *handle)
{
return(fgetc(handle->r_stream));
}

/*
** write char to stream
*/
int socket_fputc(c,handle)
int c;
socket_rec *handle;
{
return(fputc(c,handle->w_stream));
}

/*
** flush streams (read stream does not need flushing)
*/
int socket_flush(socket_rec *handle)
{
return(fflush(handle->w_stream));
}

/*
** test of eof on read stream
*/
int socket_eof(socket_rec *handle)
{
return(feof(handle->r_stream));
}

/*
** clear errors on streams
*/
void socket_clrerr(socket_rec *handle)
{
clearerr(handle->r_stream);
clearerr(handle->w_stream);
}

/*
** close streams
*/
int socket_close(socket_rec *handle)
{
fprintf(stderr,"***** closing\n");
fclose(handle->w_stream);
fclose(handle->r_stream);
close(handle->sock);
SP_free(handle);
return(0);
}



/*
------------------------------------------------------------------
connect to a socket in the Internet domain

sock=	socket for connection
host=	target hostname to be connected
port=	port of target to be connected
stream=	(pointer to) Prolog stream to be returned after connection

*** Remark:	socket stream is bi-directional
------------------------------------------------------------------
*/
int c_socket_connect_I(long int sock,char *host,long int port,SP_stream **stream)
{
struct hostent *hptr;	/*host pointer*/
struct sockaddr_in name;
SP_stream *s;
char socket_name[20];


name.sin_family=AF_INET;
if ((hptr=gethostbyname(host))==NULL)				/*get host information by name*/
	return(FAILURE);

bcopy(hptr->h_addr,&name.sin_addr,hptr->h_length);		/*copy information across*/
name.sin_port=htons(port);					/*put in port no.*/

if (connect(sock,(struct sockaddr *)&name,sizeof(name))<0)	/*connect to host*/
	return(FAILURE);

sprintf(socket_name,"%s:%d",host,port);				/*make socket name*/
s=new_socket_stream(sock,socket_name);				/*open streams on socket*/
*stream=s;
return(SUCCESS);
}


/*
** connect a socket in the UNIX domain
*/

int c_socket_connect_U(long int sock,char *path,SP_stream **stream)
{
struct sockaddr_un name;
SP_stream *s;
char socket_name[128];


name.sun_family=AF_UNIX;					/*socket family*/
if (strlen(path) > sizeof(name.sun_path)-1)			/*check path length*/
	return(FAILURE);

strcpy(name.sun_path,path);					/*copy path into name structure*/
if (connect(sock,(struct sockaddr *)&name,sizeof(name))<0)	/*try to connect*/
	return(FAILURE);

sprintf(socket_name,"<%s>",path);				/*make socket name from path*/
s=new_socket_stream(sock,socket_name);				/*create streams*/
*stream=s;
return(SUCCESS);
}



/*
------------------------------------------------------------
select active file descriptors from three lists with timeout

rs=	list of fds for read-ready testing
ws=	list of fds for write-ready testing
es=	list of fds for exception testing
sec=	time-out seconds
usec=	time-out micro-seconds
rrs=	returned list of fds ready for reading
rws=	returned list of fds ready for writing
res=	returned list of fds with exceptions
------------------------------------------------------------
*/
int c_socket_select(SP_term_ref rs,SP_term_ref ws,SP_term_ref es,long int sec,long int usec,SP_term_ref rrs,SP_term_ref rws,SP_term_ref res)
{
int max_fd=0;				/*fd counter*/
fd_set r_set;
fd_set w_set;
fd_set e_set;
struct timeval timeout;

set_fd_set(rs,&r_set,&max_fd);		/*set the read fd_set*/
set_fd_set(ws,&w_set,&max_fd);		/*set the write fd_set*/
set_fd_set(es,&e_set,&max_fd);		/*set the exception fd_set*/
if (sec==0 && usec==0)
	select(max_fd+1,&r_set,&w_set,&e_set,NULL);
else	{
	timeout.tv_sec=sec;			/*set timeout value*/
	timeout.tv_usec=usec;
	select(max_fd+1,&r_set,&w_set,&e_set,&timeout);
	}
decode_fd_set(max_fd,&r_set,rrs);
decode_fd_set(max_fd,&w_set,rws);
decode_fd_set(max_fd,&e_set,res);
return(SUCCESS);
}


/*
** set a fd_set according to a Prolog list of integers
*/


void set_fd_set(SP_term_ref list,fd_set *fset,int *max_fd)
{
SP_term_ref temp,car,cdr;
long int n;

FD_ZERO(fset);
temp=SP_new_term_ref();
car=SP_new_term_ref();
cdr=SP_new_term_ref();
SP_put_term(temp,list);				/*start with the whole list*/
while (SP_is_list(temp))			/*not the end of the list*/
	{
	SP_get_list(temp,car,cdr);		/*split into head & tail*/
	SP_get_integer(car,&n);			/*get the integer int the head*/
	FD_SET(n,fset);				/*set for testing*/
	if (n>*max_fd)				/*update max_fd if needed*/
		*max_fd=n;
	SP_put_term(temp,cdr);			/*iterate through tail*/
	}
}


/*
** decode a fd_set and construct a Prolog list of integers
*/

void decode_fd_set(int max_fd,fd_set *fset,SP_term_ref result_list)
{
SP_term_ref list,car,cdr;
int i;

list=SP_new_term_ref();
car=SP_new_term_ref();
cdr=SP_new_term_ref();

for (i=0;i<=max_fd;i++)
	if (FD_ISSET(i,fset))			/*if fd is set*/
		{
		SP_put_term(cdr,list);		/*new tail is list*/
		SP_put_integer(car,i);		/*create head*/
		SP_cons_list(list,car,cdr);	/*cons to make new list*/
		}
SP_put_term(result_list,list);			/*pass back result*/
}


/*
--------------------------------------------------
return current hostname

host=	(pointer to) string for returning hostname
--------------------------------------------------
*/
int c_hostname(char **host)
{
extern char buffer[128];
struct hostent *host_entry;
struct in_addr in;


if (gethostname(buffer,sizeof(buffer))<0)
	return(FAILURE);

/*check for fully-qualified name*/

if (strchr(buffer,'.')==NULL)
	{
	if ((host_entry=gethostbyname(buffer))==NULL)
		return(FAILURE);

	memcpy(&in.s_addr,*host_entry->h_addr_list,sizeof(in.s_addr));
	strcpy(buffer,(char *)inet_ntoa(in));
	}
*host=buffer;
return(SUCCESS);
}



